package ru.dbotthepony.kstarbound

import org.apache.logging.log4j.LogManager
import org.lwjgl.Version
import org.lwjgl.glfw.GLFW.*
import org.lwjgl.opengl.GL46.*
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.math.Matrix4f
import ru.dbotthepony.kstarbound.client.render.Camera
import ru.dbotthepony.kstarbound.client.render.ChunkRenderer
import ru.dbotthepony.kstarbound.client.render.TextAlignX
import ru.dbotthepony.kstarbound.client.render.TextAlignY
import ru.dbotthepony.kstarbound.util.Color
import ru.dbotthepony.kstarbound.util.formatBytesShort
import java.io.File

private val LOGGER = LogManager.getLogger()

var viewportWidth = 800
	private set

var viewportHeight = 600
	private set

var viewportMatrixGUI = updateViewportMatrixA()
	private set

var viewportMatrixGame = updateViewportMatrixB()
	private set

private fun updateViewportMatrixA(): Matrix4f {
	return Matrix4f.ortho(0f, viewportWidth.toFloat(), 0f, viewportHeight.toFloat(), 0.1f, 100f)
}

private fun updateViewportMatrixB(): Matrix4f {
	return Matrix4f.orthoDirect(0f, viewportWidth.toFloat(), 0f, viewportHeight.toFloat(), 1f, 100f)
}

var window = 0L
	private set

fun main() {
	LOGGER.info("Running LWJGL ${Version.getVersion()}")

	val client = StarboundClient()

	Starbound.addFilePath(File("./unpacked_assets/"))

	Starbound.initializeGame { finished, replaceStatus, status ->
		client.pushLoadingText(status, replaceStatus)
	}

	client.onTermination {
		Starbound.terminateLoading = true
	}

	while (client.renderFrame()) {
		Starbound.pollCallbacks()
	}
}

private var camera: Camera? = null
private val startupTextList = ArrayList<String>()
private var finishStartupRendering = Long.MAX_VALUE

var frameRenderTime = 1.0
	private set

val framesPerSecond get() = 1.0 / frameRenderTime

private fun loop() {
	val client = StarboundClient()
	val state = client.gl
	startupTextList.add("Initialized OpenGL context")
	camera = Camera()

	// Set the clear color
	glClearColor(0.2f, 0.2f, 0.2f, 0.2f)

	state.blend = true
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

	var chunkRenderer: ChunkRenderer? = null

	/*Starbound.onInitialize {
		val chunk = Starbound.world.getOrMakeChunk(Vector2i(2, 2))

		var x = 0
		var y = 0

		for (tile in Starbound.tilesAccess.values) {
			chunk.background[x, y + 1] = tile
			chunk.background[x++, y] = tile
			chunk.background[x, y + 1] = tile
			chunk.background[x++, y] = tile
			chunk.background[x, y + 1] = tile
			chunk.background[x++, y] = tile
			chunk.background[x, y + 1] = tile
			chunk.background[x++, y] = tile
			chunk.background[x, y + 1] = tile
			chunk.background[x++, y] = tile
			chunk.background[x, y + 1] = tile

			if (x >= 32) {
				x = 0
				y += 2
			}
		}

		val tile = Starbound.getTileDefinition("glass")

		for (x in 0 .. 32) {
			for (y in 0 .. 32) {
				chunk.foreground[x, y] = tile
			}
		}

		for (x in 4 .. 8) {
			for (y in 4 .. 8) {
				chunk.foreground[x, y] = null as TileDefinition?
			}
		}

		chunkRenderer = ChunkRenderer(state, chunk, Starbound.world)
		chunkRenderer!!.tesselateStatic()
		chunkRenderer!!.uploadStatic()
	}*/

	val runtime = Runtime.getRuntime()

	// Run the rendering loop until the user has attempted to close
	// the window or has pressed the ESCAPE key.
	while (!glfwWindowShouldClose(window)) {
		val measure = glfwGetTime()
		glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT) // clear the framebuffer

		state.matrixStack.clear(viewportMatrixGame.toMutableMatrix())
		camera?.translate(state.matrixStack.last)

		state.matrixStack.push().scale(x = 20f, y = 20f).translateWithScale(0f, 0f)
		chunkRenderer?.render()

		state.matrixStack.clear(viewportMatrixGUI.toMutableMatrix().translate(z = 2f))

		state.font.render("FPS: %.2f".format(framesPerSecond), scale = 0.4f)
		state.font.render("Mem: ${formatBytesShort(runtime.totalMemory() - runtime.freeMemory())}", x = viewportWidth.toFloat(), scale = 0.4f, alignX = TextAlignX.RIGHT)

		val thisTime = System.currentTimeMillis()

		if (startupTextList.isNotEmpty() && thisTime <= finishStartupRendering) {
			var alpha = 1f

			if (finishStartupRendering - thisTime < 1000L) {
				alpha = (finishStartupRendering - thisTime) / 1000f
			}

			state.matrixStack.push()
			state.matrixStack.translateWithScale(y = viewportHeight.toFloat())
			var shade = 255

			for (i in startupTextList.size - 1 downTo 0) {
				val size = state.font.render(startupTextList[i], alignY = TextAlignY.BOTTOM, scale = 0.4f, color = Color.SHADES_OF_GRAY[shade].copy(alpha = alpha))
				state.matrixStack.translateWithScale(y = -size.height * 1.2f)

				if (shade > 120) {
					shade -= 10
				}
			}

			state.matrixStack.pop()
		}

		glfwSwapBuffers(window) // swap the color buffers

		// Poll for window events. The key callback above will only be
		// invoked during this call.
		glfwPollEvents()
		Starbound.pollCallbacks()
		frameRenderTime = glfwGetTime() - measure
	}
}
